/*
* fcomp - a file comparison program
*
*
* Program description:
*
*	A command line has the form:
*		fcomp [-n] file1 file2
*
*	Fcomp produces a shortest sequence of instructions of the form
*		Insert the line "..." after line i.
*	or
*		Delete line i.
*	that changes file1 to file2.  If an integer n is specified, then only
*	sequences of at most n instructions are considered.
*
*
* Source files:
*
*	fcomp.c	- this file
*	lib.c	- library of C procedures
*
*
* Portability:
*
*	File1 and file2 must fit simultaneously in memory.
*/

#include <stdio.h>
#define DELETE	1
#define INSERT	2
#define MAXFILE	2000			/* maximum file length (in lines) */
#define MAXTEXT	200			/* maximum line length */
#define ORIGIN	MAXFILE			/* subscript for diagonal 0 */
#define PROG_NAME	"fcomp"		/* name of this program */

/* edit scripts are stored in linked lists */
struct edit {
	struct edit *link;		/* previous edit instruction */
	int op;				/* INSERT or DELETE */
	int line1;			/* line number in file1 */
	int line2;			/* line number in file2 */
};

char *A[MAXFILE], *B[MAXFILE];		/* lines of file1 and file2 */

main(argc, argv)
int argc;
char *argv[];
{
	int	col,			/* column number */
		d,			/* current edit distance */
		lower,			/* left-most diagonal under consideration */	
		k,			/* current diagonal */
		m,			/* number of lines in file1 */
		max_d,			/* bound on size of edit script */
		n,			/* number of lines in file2 */
		row,			/* row number */
		upper;			/* right-most diagonal under consideration */	

	/* for each diagonal, two items are saved: */
	int last_d[2*MAXFILE+1];	/* row containing the last d */
	struct edit *script[2*MAXFILE+1];     /* corresponding edit script */

	struct edit *new;
	char *malloc();

	savename(PROG_NAME);	/* for error messages */
	if (argc > 1 && argv[1][0] == '-') {
		max_d = atoi(&argv[1][1]);
		++argv;
		--argc;
	} else
		max_d = 2*MAXFILE;
	if (argc != 3)
		fatal("Two file names are required.");
	/* read in file1 and file2 */
	m = in_file(argv[1], A);
	n = in_file(argv[2], B);

	/* initialize: 0 entries in D indicate identical prefixes */
	for (row = 0; row < m && row < n && strsame(A[row], B[row]); ++row)
		;
	last_d[ORIGIN] = row;
	script[ORIGIN] = NULL;
	lower = (row == m) ? ORIGIN + 1 : ORIGIN - 1;
	upper = (row == n) ? ORIGIN - 1 : ORIGIN + 1;
	if (lower > upper) {
		puts("The files are identical.");
		exit(0);
	}

	/* for each value of the edit distance ... */
	for (d = 1; d <= max_d ; ++d) {
		/* for each relevant diagonal ... */
		for (k = lower; k <= upper; k += 2) {
			/* get space for the next edit instruction */
			new = (struct edit *) malloc(sizeof(struct edit));
			if (new == NULL)
				exceeds(d);

			/* find a d on diagonal k */
			if (k == ORIGIN-d || k != ORIGIN+d && last_d[k+1] >= last_d[k-1]) {
				/*
				* Moving down from the last d-1 on diagonal k+1
				* puts you farther along diagonal k than does
				* moving right from the last d-1 on diagonal k-1.
				*/
				row = last_d[k+1]+1;
				new->link = script[k+1];
				new->op = DELETE;
			} else {
				/* move right from the last d-1 on diagonal k-1 */
				row = last_d[k-1];
				new->link = script[k-1];
				new->op = INSERT;
			}
			/* code common to the two cases */
			new->line1 = row;
			new->line2 = col = row + k - ORIGIN;
			script[k] = new;

			/* slide down the diagonal */
			while (row < m && col < n && strsame(A[row],B[col])) {
				++row;
				++col;
			}
			last_d[k] = row;

			if (row == m && col == n) {
				/* hit southeast corner; have the answer */
				put_scr(script[k]);
				exit(0);
			}
			if (row == m)
				/* hit last row; don't look to the left */
				lower = k+2;
			if (col == n)
				/* hit last column; don't look to the right */
				upper = k-2;

		}
		--lower;
		++upper;
	}
	exceeds(d);
}

/* in_file - read in a file and return a count of the lines */
static int in_file(filename, P)
char *filename, *P[];
{
	FILE *fp, *ckopen();
	int lines = 0;
	char text[MAXTEXT], *fgets(), *strsave();

	fp = ckopen(filename, "r");
	while (fgets(text, MAXTEXT, fp) != NULL) {
		if (lines >= MAXFILE)
			fatalf("`%s' contains too many lines.", filename);
		P[lines++] = strsave(text);
	}
	fclose(fp);
	return(lines);
}

/* put_scr - print the edit script */
put_scr(start)
struct edit *start;
{
	struct edit *a, *ahead, *b, *behind, *ep;
	int change;

	/* reverse the pointers */
	ahead = start;
	ep = NULL;
	while (ahead != NULL) {
		behind = ep;
		ep = ahead;
		ahead = ahead->link;
		ep->link = behind;	/* flip the pointer */
	}

	/* print instructions */
	while (ep != NULL) {
		b = ep;
		if (ep->op == INSERT)
			printf("Inserted after line %d:\n", ep->line1);
		else {	/* DELETE */
			/* look for a block of consecutive deleted lines */
			do {
				a = b;
				b = b->link;
			} while (b!=NULL && b->op==DELETE && b->line1==a->line1+1);
			/* now b points to the instruction after the last deletion */	
			change = (b!=NULL && b->op==INSERT && b->line1==a->line1);
			if (change)
				printf("Changed ");
			else
				printf("Deleted ");
			if (a == ep)
				printf("line %d:\n", ep->line1);
			else
				printf("lines %d-%d:\n",ep->line1, a->line1);
			/* print the deleted lines */
			do {
				printf("  %s", A[ep->line1-1]);
				ep = ep->link;
			} while (ep != b);
			if (!change)
				continue;
			printf("To:\n");
		}
		/* print the inserted lines */
		do {
			printf("  %s", B[ep->line2-1]);
			ep = ep->link;
		} while (ep != NULL && ep->op == INSERT && ep->line1 == b->line1);
	}
}

/* exceeds - report that the edit distance exceeds d; terminate execution */
exceeds(d)
int d;
{
	printf("At at least %d %s inserted or deleted.\n", d,
		(d == 1) ? "line was" : "lines were" );
	exit(1);
}
