/*
* fastfind - print lines containing a given pattern string.
*
*
* Program description:
*
*	A command line has the form
*		fastfind pat [file1] [file2] ...
*	where pat is any sequence of characters.  A character pair "\n" at the
*	beginning or end of pat matches the beginning or end of a text line.
*	If no file is named, then standard input is read. If more than one file
*	is named, then each printed line is preceded by its file's name.
*
*
* Source files:
*
*	fastfind.c	- this file
*	lib.c		- library of C procedures
*
*
* Portability:
*
*	Files are read with the UNIX routines:
*
*		close(fd)
*		int fd;
*			Close the file with descriptor fd.
*
*		int open(file, mode)
*		char *file;
*		int mode;
*			Open the named file in the indicated mode.  (Mode = 0
*			means read-only.)  The returned value is a "file
*			descriptor" that is used for subsequent references to
*			the file.  A negative value is returned if the file
*			could not be opened.
*
*		int read(fd, buffer, max_chars)
*		int fd, max_chars;
*		char buffer[];
*			Read at most max_chars characters from the file with
*			file descriptor fd to the buffer.  The returned value
*			is the actual number of characters read.
*
*	File lines should be separated by '\n'. (Some systems use the pair "\r\n".)
*/

#include <stdio.h>

/*
* Two buffers of length BUFSIZ (defined in stdio.h) are arranged as a circle;
* the location just left of the current buffer is the end of the other buffer.
*/
#define TO_LEFT		buf[1-curbuf] + BUFSIZ - 1
#define DECREMENT(x)	if (x == buf[curbuf]) x = TO_LEFT; else --x
#define INCREMENT(x)	if (x == TO_LEFT) x = buf[curbuf]; else ++x
#define PROG_NAME	"fastfind"

int	curbuf,		/* current buffer */
	fd,		/* file descriptor */
	nfile,		/* number of files */
	shift[128];	/* shift table */

char	buf[2][BUFSIZ],	/* text buffers */
	*end_pat,	/* last position in the pattern */
	*file,		/* name of the file */
	*lim,		/* limit for search */
	*pat,		/* pattern */
	*pos;		/* search pointer to the text */

main(argc, argv)
int argc;
char *argv[];
{
	int i, length;
	char *p;

	savename(PROG_NAME);	/* for error messages */
	if (argc == 1)
		fatal("No pattern was given.");
	pat = argv[1];

	/* handle newline characters in the pattern */
	if (pat[0] == '\\' && pat[1] == 'n')
		*++pat = '\n';
	if ((length = strlen(pat)) == 0)
		fatal("Pattern length is zero.");
	if (length > 1 && pat[length-2] == '\\' && pat[length-1] == 'n') {
		pat[length-2] = '\n';
		pat[length-1] = '\0';
		--length;
	}
	end_pat = pat + length - 1;
	for (i = 0; i < 128; ++i)
		shift[i] = length;
	for (p = pat; *p != '\0'; ++p)
		shift[*p & 0177] = --length;
	if ((nfile = argc - 2) == 0) {
		fd = 0;			/* standard input */
		scan();
	} else
		/* for each specified file */
		for (i = 2; i < argc; ++i) {
			file = argv[i];
			if ((fd = open(file, 0)) < 0)
				fprintf(stderr, "%s: Cannot open %s.\n",
					PROG_NAME, file);
			else {
				scan();
				close(fd);
			}
		}
	exit(0);
}

/* scan - find lines in file that contain the pattern string */
scan()
{
	int increment;

	buf[1][BUFSIZ-1] = '\n';	/* in case the first line matches */
	curbuf = 1;
	lim = pos = buf[1];	/* force an immediate call to fill_buffer() */

	for ( ; ; ) {
		/*
		* Pos points to a text position that might end an occurrence
		* pat. If that character differs from the last character of pat,
		* then pos is shifted to the next position that might yield a
		* match.  The shifting stops when pos reaches the end of the
		* buffer or an instance of the last pattern character.
		*/ 
		while (pos < lim && (increment = shift[*pos & 0177]) > 0)
			pos += increment;
		if (pos < lim) {	/* shifting ended with pos in buffer */
			if (is_match())
				print_line();
			++pos;
		}
		/* else past end of buffer; fill the other buffer */
		else if (fill_buffer() == EOF)
			break;
	}
}

/*
* fill_buffer - fill other buffer; point pos to first char read; point lim just
* beyond the last character read; return EOF at end of file.
*/
int fill_buffer()
{
	curbuf = 1 - curbuf;
	pos = buf[curbuf];
	if ((lim = pos + read(fd, pos, BUFSIZ)) == pos)
		return(EOF);
	return(!EOF);
}

/* is_match - tell if a copy of the pattern ends at pos */
int is_match()
{
	char *t = pos, *p = end_pat;	/* already know that *t == *p */

	while (--p >= pat) {
		DECREMENT(t);
		if (*p != *t)
			return(0);
	}
	return(1);
}

/* print_line - print the line pointed to by pos; move pos to the end of line */
print_line()
{
	char *t;

	if (nfile > 1)
		printf("%s:", file);

	/* find the start of the line */
	if (*pos == '\n')
		DECREMENT(pos);
	for (t = pos; *t != '\n'; )
		DECREMENT(t);

	/* print the portion of the line before the match */
	while (t != pos) {
		INCREMENT(t);
		putchar(*t);
	}

	/* print the portion of the line after the match */
	while (*pos != '\n') {
		if (++pos >= lim && fill_buffer() == EOF)
			break;
		putchar(*pos);
	}
}
