/*
* macmod.c - macro symbol table
*
*
* Entry points:
*
*	macin(name, defn)
*	char *name, *defn;
*		Add the name and definition to the symbol table.  If the name
*		is already defined, then discard the old definition.
*
*	macout(from, to, max_chars)
*	char *from, *to;
*	int max_chars;
*		Copy the first argument to the second, expanding macro names.
*		Macro names begin with a letter (lower or upper case) and
*		contain letters, digits and the underscore (_) character.
*		Macro names used in macro definitions are expanded recursively.
*		Putting more than max_chars characters into the target string
*		causes a fatal error.
*
*
* Implementation:
*
*	Macros are stored in a simple linked list.
*/

#include <stdio.h>
#include <ctype.h>

struct macro {
	char *name;
	char *def;
	struct macro *next;
};

/* start points to the list of macros, which is initially empty */
static struct macro *start = (struct macro *) NULL;

/* macin - install a macro name and definition */
macin(name, defn)
char *name, *defn;
{
	struct macro *mp, *search();
	char *ckalloc(), *strsave();

	/* if the macro is not yet defined */
	if ((mp = search(name)) == NULL) {
		/* allocate storage for a node */
		mp = (struct macro *) ckalloc(sizeof(struct macro));
		/* install the name */
		mp->name = strsave(name);
		/* insert the node at the start of the linked list */
		mp->next = start;
		start = mp;
	} else	/* if the macro is being redefined */
		/* free the old definition's storage */
		free(mp->def);
	/* in either case, install the definition */
	mp->def = strsave(defn);
}

/* macout - expand macro names in a string; do not exceed max_chars */
macout(from, to, max_chars)
char *from, *to;
int max_chars;
{
	char *junk, *expand(), *limit;

	limit = &to[max_chars - 1];	/* last permissible address */
	junk = expand(from, to, limit);
}

/*
* expand - recursive procedure to expand macros; do not exceed limit;
*	return pointer to next position in the output string.
*
* The basic algorithm is as follows.  Copy the string beginning at "from"
* to the string beginning at "to".  Check for substrings that are potential
* macro names.  Whenever a potential macro name has been copied, look it
* up in the macro table.  If it is found, then call expand() to overwrite
* the copied macro name with the macro's definition.
*/
static char *expand(from, to, limit)
char *from, *to, *limit;
{
	struct macro *mp, *search();
	static int depth = 0;		/* depth of recursion */
	char *potential;

	if (++depth > 100)
		fatal("Infinite recursion while expanding macros.");
	for ( ; ; ) {
		/* find the start of a potential macro name */
		while (!isalpha(*from))
			/* copy characters; check for end of string */
			if ((*to++ = *from++) == '\0') {
				--depth;
				return(to - 1);
			} else if (to >= limit)
				fatal("Line overflow during macro expansion.");
		/* point to the start of a potential macro name */
		potential = to;
		/* find end of potential macro name */
		while (isalnum(*from) || *from == '_') {
			*to++ = *from++;
			if (to >= limit)
				fatal("Line overflow during macro expansion.");
		}
		/* mark end of potential macro name to prepare for search */
		*to = '\0';
		/* if it is a macro name, then expand it */
		if ((mp = search(potential)) != NULL)
			to = expand(mp->def, potential, limit);
	}
}

/* search - look for macro, return pointer to structure */
static struct macro *search(name)
char *name;
{
	struct macro *mp;

	/* scan linked list of macros */
	for (mp = start; mp != NULL; mp = mp->next)
		/* if the name is found */
		if (strsame(name, mp->name))
			/* return pointer to the structure */
			return(mp);
	/* if the name is not found, return NULL */
	return(NULL);
}
