/*
* macro - rudimentary macro processor
*
*
* Program description:
*
*	A command line has the form:
*		macro filename
*	Macro definition lines begin with the eight characters "#define<space>".
*	Spaces and tab characters preceding the macro name, or falling between
*	the name and the definition, are discarded and the trailing newline
*	character is removed.
*
*
* Source files:
*
*	lib.c		- library of C procedures
*	macmod.c	- macro symbol table
*	macro.c		- this file
*
*
* External procedure calls:
*
*	macin(name, defn)
*	char *name, *defn;
*		Add the name and definition to the symbol table.  If the name
*		is already defined, then discard the old definition.
*
*	macout(from, to, max_chars)
*	char *from, *to;
*	int max_chars;
*		Copy the first argument to the second, expanding macro names.
*		Macro names begin with a letter (lower or upper case), and
*		contain letters, digits and the underscore (_) character.
*		Macro names used in macro definitions are expanded recursively.
*		Putting more than max_chars characters into the target string
*		causes a fatal error.
*
*/

#include <stdio.h>
#include <ctype.h>
#define MAXTEXT 200
#define PROG_NAME "macro"

static char *name, *defn;

main(argc, argv)
int argc;
char *argv[];
{
	FILE *ckopen(), *fp;
	char *fgets(), text[MAXTEXT], text2[MAXTEXT];

	savename(PROG_NAME);	/* for error messages */
	if (argc != 2)
		fatal("Expecting exactly one file name.");
	fp = ckopen(argv[1], "r");

	/* for each input line */
	while (fgets(text, MAXTEXT, fp) != NULL)
		/* if the line begins with the eight characters "#define<space>" */
		if (is_define(text)) {
			/* split the rest of the line into a name and a definition */
			split(text);
			/* remember the name and the definition */
			macin(name, defn);
		} else {
			/* replace each macro name in the line by its definition */
			macout(text, text2, MAXTEXT);
			/* print the resulting line */
			fputs(text2, stdout);
		}
	fclose(fp);
	exit(0);
}

/* is_define - tell if the first 8 characters are "#define<space>" */
static int is_define(text)
char *text;
{
	char *s = "#define ";

	while (*s != '\0')
		if (*s++ != *text++)
			return(0);
	return(1);
}

/* split - split #define line into a name and definition */
static split(text)
char *text;
{
	char *end;

	/* skip whitespace after "#define " */
	for (name = text+strlen("#define"); *name == ' ' || *name == '\t'; ++name)
		;
	/* macro name extends to first whitespace character */
	for (end = name; !isspace(*end); ++end)
		;
	*end = '\0';	/* mark the end of the name */

	/* skip whitespace to find the definition */
	for (defn = end + 1; *defn == ' ' || *defn == '\t'; ++defn)
		;
	/* strip the newline from the end of the definition */
	for (end = defn; *end != '\n' && *end != '\0'; ++end)
		;
	*end = '\0';

}
