/*
* FSMmod.c - build and save the finite-state machine
*
*
* Entry points:
*
*	state(i, type_ptr, value_ptr, lson_ptr, rson_ptr)
*	int i, *type_ptr, *value_ptr, *lson_ptr, *rson_ptr;
*		Return the information associated with state i.  For labeled
*		states, the type and value give the literal associated with the
*		state.  For NIL states, the type and value are NIL.  The states
*		adjacent to state i are returned through the last two arguments.
*		The start state is the left son of the pseudostate i = 0.
*
*	int treelit(type, value)
*	int type, value;
*		Create a one-state machine labeled by type and value.  Return
*		the number of the state.
*
*	int treeop(op, left, right)
*	int op, left, right;
*		Apply the construction for op (CAT, |, + or *) to submachines
*		with the given start states.  Return the start state of the new
*		machine.
*
* Implementation:
*
*	States are stored in "parallel" arrays.
*/
#include "match.h"

static int
	last[2*EXPRMAX],	/* final state for machine starting here */
	lson[2*EXPRMAX],	/* first adjacent state or NIL */
	next_state = 0,		/* states are numbered 0, 1, 2, ... */
	rson[2*EXPRMAX],	/* second adjacent state or NIL */
	type[2*EXPRMAX],	/* label (for literals) or NIL */
	value[2*EXPRMAX];	/* literal's value (for a labeled state) */

/* state - return the attributes of a state */
state(i, type_ptr, value_ptr, lson_ptr, rson_ptr)
int i, *type_ptr, *value_ptr, *lson_ptr, *rson_ptr;
{
	if (i < 0 || i >= next_state)
		fatal("Impossible state number.");
	*type_ptr = type[i];
	*value_ptr = value[i];
	*lson_ptr = lson[i];
	*rson_ptr = rson[i];
}

/* treelit - build the one-state machine for a literal */
int treelit(typ, val)
int typ, val;
{
	int start_state;

	if (next_state == 0)	/* initialize the pseudostate */
		make_state(NIL, NIL, NIL, NIL, NIL);
	start_state = next_state;
	make_state(typ, val, NIL, NIL, next_state);
	return(lson[0] = start_state);
}

/* treeop - build the machine for the subtree rooted at an operator node */
int treeop(op, left, right)
int op, left, right;
{
	int start_state;

	switch (op) {
		case CAT:
			/* join the left machine to the right machine */
			start_state = left;
			lson[last[left]] = right;
			last[start_state] = last[right];
			break;
		case '|':
			/* create a state pointing to the two start states */
			start_state = next_state;
			make_state(NIL, NIL, left, right, next_state+1);
			/* point the two final states to a new final state */
			lson[last[left]] = lson[last[right]] = next_state;
			make_state(NIL, NIL, NIL, NIL, NIL);
			break;
		case '+':
		case '*':
			/* create a new state following the old final state */
			start_state = (op == '+') ? left : next_state;
			lson[last[left]] = next_state;
			make_state(NIL, NIL, left, next_state+1, NIL);
			/* create a new final state */
			last[start_state] = next_state;
			make_state(NIL, NIL, NIL, NIL, NIL);
			break;
	}
	return(lson[0] = start_state);
}

/* make_state - create a state with specified attributes */
static make_state(typ, val, left, right, final)
int typ, val, left, right, final;
{
	type[next_state] = typ;
	value[next_state] = val;
	lson[next_state] = left;
	rson[next_state] = right;
	last[next_state++] = final;
}
