/*
* Smod.c - convert between sets S and S_numbers
*
*
* Entry points:
*
*	int final(S_nbr)
*	int S_nbr;
*		Tell if the set S with the indicated S_number is final.
*
*	int S_number(S)
*	int S[];
*		Return the S_number of the given list of states.  The list must
*		be ended with -1.  If this is the first encounter with S, then
*		(i) S is stored and a new S_number is allocated for it, (ii) the
*		set T of labeled states that are reachable from S is found and
*		stored, and (iii) it is determined whether S is final.
*
*	int *T_first(S_nbr)
*	int S_nbr;
*		Return a pointer to array of labeled states that are reachable
*		from the designated S. The list is ended with -1.
*
*
* Implementation:
*
*	The members of sets S and T are stored in the array entry[]. Each list
*	of states is ended with -1.
*/

#include "match.h"
#define LIMIT 2000		/* limits total sizes of sets S and T */

static int entry[LIMIT];	/* entries of sets S and T; -1 = end */
static int is_final[128];	/* tell if the i-th S is final */
static int *next_loc = entry;	/* next free location for storing a set S or T */
static int next_S = 0;		/* next free S_number */
static int R[2*EXPRMAX];	/* holds states reachable from the current S */
static int *R_next;		/* points to next free position in R */
static int *S_start[128];	/* points to the first state in the i-th S */
static int *T_start[128];	/* points to the first state in the i-th T */

/* final - tell whether S contains a final state */
int final(S_nbr)
int S_nbr;
{
	return(is_final[S_nbr]);
}

/* S_number - convert a list of states into an S_number */
int S_number(S)
int *S;
{
	int nbr, *S_copy;


	/* make a copy of S */
	S_copy = next_loc;
	do
		save_state(*S);
	while (*S++ != -1);

	/* sort the states to simplify checking sets for equality */
	sort_states(S_copy);

	/* search sequentially through the allocated S_numbers */
	for (nbr = 0; nbr < next_S; ++nbr)
		if (same_states(S_copy, S_start[nbr])) {
			/* found S; reset next_loc to discard the copy */
			next_loc = S_copy;
			return(nbr);
		}

	/* search failed; assign a new S_number, compute T and is_final[] */
	if (next_S >= 128)
		fatal("Too many sets of states.");
	S_start[next_S] = S_copy;
	T_start[next_S] = next_loc;
	make_T(S_start[next_S]);
	return(next_S++);
}

/* T_first - convert an S_number into the associated set T */
int *T_first(S_nbr)
int S_nbr;
{
	return (T_start[S_nbr]);
}

/*
* make_T - compute T and is_final[] for the state set S
*
* T is the set of labeled states that are reachable from S (that is, the
* states that can be reached from a state in S by a non-void path that
* crosses only NIL nodes).  S is final if either (i) S contains a state that
* has only NIL edges out of it or (ii) there is a NIL state that is reachable
* from S and which has only NIL edges out of it.
*/
static make_T(S)
int S[];
{
	int lson, rson, type, value;

	R_next = R;	/* empty the list of reached states */

	for ( ; *S != -1; ++S) {
		state(*S, &type, &value, &lson, &rson);
		if (lson == NIL && rson == NIL)
			/* condition (i) for S to be final */
			is_final[next_S] = 1;
		reachable(lson);
		reachable(rson);
	}
	save_state(-1);
}

/* reachable - process a state that is reachable from S */
static reachable(sta)
int sta;
{
	int lson, *r, rson, type, value;

	if (sta == NIL)
		return;
	/* avoid redundant computation; check if state was reached earlier */
	for (r = R; r < R_next; ++r)
		if (*r == sta)
			return;

	/* marke state as reached and process it */
	*R_next++ = sta;
	state(sta, &type, &value, &lson, &rson);

	/* if the state is a NIL state ... */
	if (type == NIL) {
		if (lson == NIL && rson == NIL)
			/* condition (ii) for S to be final */
			is_final[next_S] = 1;
		/* ... then its offspring are reachable */
		reachable(lson);
		reachable(rson);
	} else	/* its a labeled state, so its in T */
		save_state(sta);
}

/* sort_states - bubble sort the states in S */
static sort_states(S)
int S[];
{
	int i, j, temp;

	for (i = 0; S[i] != -1; ++i)
		for (j = i+1; S[j] != -1; ++j)
			if (S[i] > S[j]) {
				temp = S[i];
				S[i] = S[j];
				S[j] = temp;
			}
}

/* same_states - check for identity of two lists of states */
static int same_states(s, t)
int *s, *t;
{
	for ( ; *s == *t; ++s, ++t)
		if (*s == -1)
			return(1);
	return(0);
}

/* save_state - save a state; die if no room is left */
static save_state(state)
int state;
{
	if (next_loc - entry >= LIMIT)
		fatal("Too many states.");
	*next_loc++ = state;
}
