/*
* match - print lines that match a regular expression
*
*
* Program description:
*
*	A command line has the form
*		match [-m] [-n] pat [file1] [file2] ...
*	where pat is any regular expression.  If no file is named, then standard
*	input is read.  If more than one file is named, then each printed line
*	is preceded by its file's name.  The -m flag asks for lines that don't
*	match the pattern.  The -n flag asks that line numbers be printed.
*
*
* Source files:
*
*	CCLmod.c	- store expanded character classes
*	FSMmod.c	- build and save the finite-state machine
*	lib.c		- a library of C procedures
*	match.c		- this file; main program and scanning procedure
*	match.h		- macros for the match program
*	parse.c		- parse a regular expression
*	Smod.c		- convert between sets S and S_numbers
*	token.c		- break the regular expression into tokens
*
*
* Portability:
*
*	If the command interpreter attaches a special meaning to any of the
*	regular expression meta-characters
*		\ . [ ] ! - | + * ( )
*	then it may be desirable to employ other characters.  The appropriate
*	macros in match.h can be redefined.
*
*
* External procedure calls:
*
*	cclmem(c, class_id)				.. file CCLmod.c
*	char c;
*	int class_id;
*		Tell if c is in the identified character class.
*
*	int final(S_nbr)				.. file Smod.c
*	int S_nbr;
*		Tell if the set S with the indicated S_number is final.
*
*	parse(pattern)					.. file parse.c
*	char *pattern;
*		Parse the given regular expression and build an equivalent
*		finite-state machine.
*
*	int S_number(S)					.. file Smod.c
*	int S[];
*		Return the S_number of the given list of states.  The list
*		must be ended with -1.  If this is the first encounter with S,
*		then (i) S is stored and a new S_number is allocated for it,
*		(ii) the set T of labeled states that are reachable from S is
*		found and stored, and (iii) it is determined whether S is final.
*
*	state(i, type_ptr, val_ptr, lson_ptr, rson_ptr)	.. file FSMmod.c
*	int i, *type_ptr, *val_ptr, *lson_ptr, *rson_ptr;
*		Return the information associated with state i.  For labeled
*		states, the type and value give the literal associated with the
*		state.  For NIL states, the type and value are NIL.  The states
*		adjacent to state i are returned through the last two arguments.
*		The start state is the left son of the pseudostate i = 0.
*
*	int *T_first(S_nbr)				.. file Smod.c
*	int S_nbr;
*		Return a pointer to array of labeled states that are reachable
*		from the designated S. The list is ended with -1.
*/

#include "match.h"
#define MAXTEXT	200
#define YES	1
#define NO	0

static int mflag = NO, nfiles, nflag = NO;

main(argc, argv)
int argc;
char *argv[];
{
	static int S[] = {0, -1};
	int i, j;
	char pattern[EXPRMAX];

	savename(PROG_NAME);	/* for error messages */

	for (i = 1; i < argc && argv[i][0] == '-'; ++i)
		for (j = 1; argv[i][j] != '\0'; ++j)
			switch (argv[i][j]) {
				case 'm':
					mflag = YES;
					break;
				case 'n':
					nflag = YES;
					break;
				default:
					fatal("Permissible flags are m and n.");
			}
	if (i == argc)
		fatal("No pattern was specified.");
	/* augment pattern so it matches entire lines */
	if (strlen(argv[i]) > EXPRMAX-7)
		fatal("The regular expression is too long.");
	sprintf(pattern,"%c%c%c%s%c%c%c", ANY, STAR, L_GROUP, argv[i],
		R_GROUP, ANY, STAR);
	/* parse the pattern; build the finite-state machine */
	parse(pattern);
	/* record the initial state set */
	S_number(S);
	if ((nfiles = argc - i - 1) == 0)
		scan("");
	else
		while (++i < argc)
			scan(argv[i]);
	exit(0);
}

/* scan - find lines in file that match the pattern */
static scan(file)
char *file;
{
	FILE *fp, *ckopen();
	int c, i, line_number = 0, nbr_S, S[EXPRMAX], *T_first(), *T_ptr, try;
	char *fgets(), text[MAXTEXT+1], *p;

	/* static variables are guaranteed to start off as 0 */
	static char new_S[128][128];

	if (nfiles == 0)
		fp = stdin;
	else
		fp = ckopen(file, "r");

	text[0] = '\n';	/* implicit newline at the start of each input line */

	while (fgets(text+1, MAXTEXT, fp) != NULL) {
		++line_number;
		/* begin scanning in the initial state set */
		nbr_S = 0;
		/* loop over the input characters c */
		for (p = text; *p != '\0'; ++p) {
			c = *p & 0177;	/* thwart sign-extension */
			if ((try = new_S[nbr_S][c]) > 0)
				nbr_S = try;
			else {
				/*
				* This is a new combination of S and c.
				* For T = set of states that follow a state in S,
				* compute S = set of states in T that match c.
				*/
				for (i = 0, T_ptr = T_first(nbr_S); *T_ptr != -1; ++T_ptr)
					if (matches(c, *T_ptr))
						S[i++] = *T_ptr;
				S[i] = -1;
				/* save the results of all this work */
				nbr_S = new_S[nbr_S][c] = S_number(S);
			}
		}

		if (final(nbr_S) != mflag) {
			if (nfiles > 1)
				printf("%s: ",file);
			if (nflag == YES)
				printf("%d: ", line_number);
			fputs(text+1, stdout);
		}
	}
	if (fp != stdin)
		fclose(fp);
}

/* matches - tell if character c matches the label on state s */
static matches(c, s)
int c, s;
{
	int junk1, junk2, type, value;

	state(s, &type, &value, &junk1, &junk2);
	switch (type) {
		case '.':
			return (1);
		case CHAR:
			return (c == value);
		case CCL:
			return ( cclmem((char)c, value) );
		case NCCL:
			return ( !cclmem((char)c, value) );
		default:
			fatal("matches(): Cannot happen.");
	}
}
