/*
* parse.c - parse a regular expression
*
*
* Entry point:
*
*	parse(pattern)
*	char *pattern;
*		Parse the given regular expression.
*
*
* External procedure calls:
*
*	get_token(val_ptr)				.. file token.c
*	int *val_ptr;
*		Return the next token, where the regular expression is divided
*		into the tokens:
*			CHAR  CCL  NCCL  .  |  +  *  (  )  '\0'
*		For tokens CHAR, CCL and NCCL, get_token()'s argument is used
*		to return an additional value, namely the specific character
*		for CHAR, or the character class ID for CCL or NCCL.
*
*	tokinit(pattern)				.. file token.c
*	char *pattern;
*		Prepare to break the given regular expression into tokens.
*
*
*	int treelit(type, value)			.. file FSMmod.c
*	int type, value;
*		Create a one-node expression tree for a literal with the given
*		type and value.  Return a number that uniquely identifies the tree.
*
*	int treeop(type, lson, rson)			.. file FSMmod.c
*	int type, lson, rson;
*		Create an expression tree with root node of the given type
*		and whose left and right subtrees have the given identifying
*		numbers.  Return a number that uniquely identifies the tree.
*/

#include "match.h"

static int token;	/* the current token */
static int value;	/* the value associated with the current token */

/* parse - entry point to build an expression tree */
parse(pattern)
char *pattern;
{
	tokinit(pattern);
	token = get_token(&value);
	expression();
	if (token != '\0')
		fatal("Syntax error in pattern.");
}

/*
* expression - build the expression tree for the expression beginning at token;
*		raise token just past the expression.
*/
static int expression()
{
	int subtree;

	/* break expression into terms; get expression trees; join with | nodes */
	subtree = term();
	while (token == '|') {
		token = get_token(&value);
		subtree = treeop((int)'|', subtree, term());
	}
	return(subtree);
}

/*
* term - build the expression tree for the term beginning at token;
*		raise token just past the term.
*/
static int term()
{
	int subtree;

	/* break term into factors; get expression trees; join with CAT nodes */
	subtree = factor();
	while (token == '(' || LITERAL(token))
		subtree = treeop(CAT, subtree, factor());
	return(subtree);
}

/*
* factor - build the expression tree for the factor beginning at token;
*		raise token just past the factor.
*/
static int factor()
{
	int subtree;

	/* get subtree for parenthesized expression or literal */
	if (token == '(') {
		token = get_token(&value);
		subtree = expression();
		if (token != ')')
			fatal ("Unbalanced parentheses in regular expression.");
	} else if LITERAL(token)
		subtree = treelit(token, value);
	else
		fatal("Syntax error in the regular expression.");
	token = get_token(&value);

	/* add nodes for subsequent unary operators */
	while (token == '*' || token == '+') {
		subtree = treeop(token, subtree, 0);
		token = get_token(&value);
	}
	return(subtree);
}
