/*
* token.c - break a regular expression into tokens
*
*
* Entry points:
*
*	int get_token(val_ptr)
*	int *val_ptr;
*		Return the next token, where the regular expression is divided
*		into the tokens:
*			CHAR  CCL  NCCL  .  |  +  *  (  )  '\0'
*		For tokens CHAR, CCL and NCCL, get_token()'s argument is used
*		to return an additional value, namely the specific character
*		for CHAR, or the character class ID for CCL or NCCL.
*
*	tokinit(pattern)
*	char *pattern;
*		Prepare to break the given regular expression into tokens.
*
*
* External procedure calls:
*
*	cclin(c)					.. file CCLmod.c
*	char c;
*		Store c in the current character class.  Character classes
*		(and negated character classes) are saved by calling cclnew()
*		to get a character class ID, then storing the characters one
*		at a time by calling cclin(char).
*
*	int cclnew()					.. file CCLmod.c
*		Return a character class ID, which is a small integer used
*		for later identification of the character class.
*/

#include "match.h"

static char *pscan;		/* scans pattern */

/* tokinit - initialize for producing tokens */
tokinit(pattern)
char *pattern;
{
	pscan = pattern;
}

/* get_token - return token (and perhaps value) */
int get_token(val_ptr)
int *val_ptr;
{
	int c;

	/* Replace symbolic meta-characters by the standard ones. */
	switch(c = *pscan++) {
		case ANY:
			c = '.';
			return(c);
		case CCL: 
			return(bldccl(val_ptr));
		case L_GROUP:
			c = '(';
			return(c);
		case OR:
			c = '|';
			return(c);
		case PLUS:
			c = '+';
			return(c);
		case R_GROUP:
			c = ')';
			return(c);
		case STAR:
			c = '*';
			return(c);
		case '\0':
			return (c);
		case ESCAPE:
			if ((c = *pscan++) == '\0')
				return(c);
			if (c == 'n')
				c = '\n';
			/* a \ has been stripped from the front of a token --
			 *  fall through to the code for character constants */
		default:
			*val_ptr = c;
			return(CHAR);
	}
}

/*
* bldccl - build a character class or negated character class
*
* The character class ID is returned through the argument; the function's value
* is CCL or NCCL.
*/
static int bldccl(id)
int *id;
{
	int token;
	static char digits[] = "0123456789";
	static char lowers[] = "abcdefghijklmnopqrstuvwxyz";
	static char uppers[]  = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
	char ch, *index();

	/* assign an ID to the character class */
	*id = cclnew();
	/* distinguish between a CCL and a NCCL */
	if (*pscan == NCCL) {
		token = NCCL;
		pscan++;
	} else
		token = CCL;
	/* a leading ] should be treated as a character constant */
	if (*pscan == CCL_END)
		cclin(*pscan++);
	/* scan the list of characters */
	for ( ; *pscan != CCL_END && *pscan != '\0'; ++pscan)
		/* look for a newline character, specified as "\n" */
		if (pscan[0] == ESCAPE && pscan[1] != '\0')
			if ((ch = *++pscan) == 'n')
				cclin('\n');
			else
				cclin(ch);
		else if (pscan[1] != RANGE || pscan[2] == CCL_END)
			cclin(*pscan);
		/* remaining cases handle a - between other characters */
		else if (index(digits,*pscan) > 0 && index(digits,pscan[2]) > 0)
			expand(digits);
		else if (index(lowers,*pscan) > 0 && index(lowers,pscan[2]) > 0)
			expand(lowers);
		else if (index(uppers,*pscan) > 0 && index(uppers,pscan[2]) > 0)
			expand(uppers);
		else
		/* characters around '-' were incomparable; its a plain '-' */
			cclin(*pscan);
	if (*pscan++ != CCL_END)
		fatal("Character class not properly terminated.");
	return(token);
}

/* expand - expand the range of characters pscan[0]-pscan[2] */
static expand(chars)
char *chars;
{
	char *from, *index(), *k, *to;

	from = index(chars, *pscan++);
	to = index(chars, *++pscan);
	for (k = from; k <= to; ++k)
		cclin(*k);
}
