/*
* bldint.c - build internal form of the UPDATE file
*
*
* Entry point:
*
*	int bldint()
*		Read the UPDATE file and record the information.  Return
*		a count of the syntax errors in UPDATE.
*
*
* External procedure calls:
*
*	cmdin(file, command)				.. file depmod.c
*	char *file, *command;
*		Record a command of a file.
*
*	macin(name, definition)				.. file macmod.c
*	char *name, *definition;
*		Record a macro name and definition.
*
*	macout(line1, line2, maxchars)			.. file macmod.c
*	char *line1, *line2;
*	int maxchars;
*		Expand macros while copying line1 to line2.  At most
*		maxchars characters are copied.
*
*	prein(file, precursor)				.. file depmod.c
*	char *file, *precursor;
*		Record a precursor of a file.
*/

#include "update.h"

static FILE *U_file;		/* UPDATE file */
static int line_nbr = 0;	/* current position in UPDATE file */

/*
* bldint - classify UPDATE lines; return error count
*
* Macro definition lines begin with '#'; commands are preceded by a tab
* character; the rest are dependency lines.
*/
int bldint()
{
	FILE *ckopen();
	int errors = 0;
	char dependents[CMD_SIZE], line[CMD_SIZE], line2[CMD_SIZE];
	/*
	* dependents[] holds the names on the left of the most recent dependency
	* line, separated by '\0' characters and terminated with two '\0's.  The
	* dependents are determined by blddep() and used by bldcmd().
	*/

	U_file = ckopen(U_NAME, "r");  /* U_NAME is "UPDATE" */
	dependents[0] = '\0';
	while (inline(line) != EOF)
		switch (*line) {
			case '#':
				errors += bldmac(line);
				break;
			case '\t':
				errors += bldcmd(line, dependents);
				break;
			default:
				macout(line, line2, CMD_SIZE);
				errors += blddep(line2, dependents);
		}
	fclose(U_file);
	return(errors);
}

/* bldmac - process a macro definition line; return error count */
static int bldmac(line)
char *line;
{
	char *d, *n = line + 7;	/* 7 = length of "#define" */

	if (!isspace(*n))
		return(improper("macro definition"));
	*n++ = '\0';
	if (!strsame(line, "#define"))
		return(improper("macro definition"));
	/* find start of name */
	while (isspace(*n))
		++n;
	if (*n == '\0')
		return(improper("macro definition"));
	/* find end of name */
	for (d = n+1; !isspace(*d) && *d != '\0'; ++d)
		;
	if (isspace(*d))
		*d++ = '\0';
	/* find start of definition */
	while (isspace(*d))
		++d;
	macin(n, d);
	return(0);
}

/* bldcmd - process a command; return error count */
static int bldcmd(line, dependents)
char *line, *dependents;
{
	if (dependents[0] == '\0')
		return(improper("command"));
	/* strip blanks and tabs from front of command */
	while (isspace(*line))
		++line;
	/* for each dependent file ... */
	for ( ; *dependents != '\0'; dependents += strlen(dependents)+1)
		cmdin(dependents, line);
	return(0);
}

/* blddep - process a dependency line; return error count */
static int blddep(line, dependents)
char *line, *dependents;
{
	char *e, *f = dependents, *p = line;

	/* copy names on left of "->", separate with '\0'. */
	while (*p != '\0' && (p[0] != '-' || p[1] != '>'))
		if (isspace(*f++ = *p++)) {
			f[-1] = '\0';
			while (isspace(*p))
				++p;
		}
	f[1] = f[0] = '\0';
	if (*p == '\0')
		return(improper("dependency line"));

	p += 2;			/* skip "->" */

	/* for each name on the right of "->", record all precursor relations */
	for ( ; ; ) {
		/* find start of name */
		while (isspace(*p))
			++p;
		if (*p == '\0')
			break;
		/* find end of name */
		for (e = p+1; !isspace(*e) && *e != '\0'; ++e)
			;
		if (isspace(*e))
			*e = '\0';
		else
			e[1] = '\0';	/* terminate on next iteration */
		/* for each name to the left of "->" ... */
		for (f = dependents; *f != '\0'; f += strlen(f) + 1)
			prein(f, p);	/* record the dependency */
		p = e + 1;
	}
	return(0);
}

/*
* inline - input an UPDATE line; return its last character ('\n' or EOF)
*
* Comments are removed; continued lines (those ending with '\') are collected;
* the terminating '\n' is removed; lines containing only whitespace are ignored.
*/
static int inline(line)
char *line;
{
	int c = U_char();
	char *line_start = line, *s;

	while (c != EOF && line - line_start < CMD_SIZE)
		switch (c) {
			case '/':	/* possible comment */
				if ((c = U_char()) == '*') {
					skip_comment();
					c = U_char();
				} else
					*line++ = '/';
				break;
			case '\\':	/* possible continued line */
				if ((c = U_char()) == '\n') {
					/* back up over white space */
					while (line > line_start && isspace(line[-1]))
						--line;
					*line++ = ' ';
					while (isspace(c = U_char()))
						;
				} else
					*line++ = '\\';
				break;
			case '\n':
				for (s = line_start; s < line; ++s)
					if (!isspace(*s)) {
						*line = '\0';
						return(c);
					}
				/* ignore lines containing only white space */
				line = line_start;
				c = U_char();
				break;
			default:
				*line++ = c;
				c = U_char();
		}
	if (line - line_start >= CMD_SIZE)
		fatalf("%s line exceeds maximum length", U_NAME);
	*line = '\0';
	return(c);
}

/* U_char - get character from UPDATE file; count lines */
static int U_char()
{
	int c;

	if ((c = getc(U_file)) == '\n')
		++line_nbr;
	return(c);
}

/* skip_comment - discard a comment */
static skip_comment()
{
	int new_c = '\0', old_c = '\0';

	while (old_c != '*' || new_c != '/') {
		old_c = new_c;
		if ((new_c = U_char()) == EOF)
			fatalf("EOF in %s comment", U_NAME);
	}
}

/* improper - report syntax error in UPDATE line; return error increment */
static int improper(type)
char *type;
{
	fprintf(stderr, "Improper %s at line %d of %s.\n",
		type, line_nbr, U_NAME);  /* U_NAME is "UPDATE" */
	return(1);
}
