/*
* depmod.c - data structure for dependencies
*
*
* Entry points:
*
*	cmdin(file, command)
*	char *file, *command;
*		Record a command of a file.
*
*	char *cmdout(file, n)
*	char *file;
*	int n;
*		Return file's n-th command (NULL if there aren't n commands).
*
*	prein(file, precursor)
*	char *file, *precursor;
*		Record a precursor of a file.
*
*	char *preout(file, n)
*	char *file;
*	int n;
*		Return file's n-th precursor (NULL if there aren't n precursors).
*
*	statin(file, status)
*	char *file;
*	int status;
*		Record the status of a file.
*
*	int statout(file)
*	char *file;
*		Return file's current status.
*
*	char *topname()
*		Return the first file name to be recorded.
*
*
* Implementation:
*
*	File information is stored in a simple linked list of "filenodes", where
*	each file has two lists (precursors and commands) of linked "listnodes".
*/

#include "update.h"

struct listnode {
	char *string;
	struct listnode *next;
};

struct filenode {
	char *name;			/* file's name */
	struct listnode *pres;		/* precursors */
	int pre_nbr;			/* number of precursors */
	struct listnode *cmds;		/* commands */
	int cmd_nbr;			/* number of commands */
	int status;			/* file's status */
	struct filenode *link;		/* link to the next file */
};

static char *first_file = NULL;		/* first file name to be recorded */
static struct filenode *begin = NULL;	/* points to the list of files */

/* cmdin - record a command of a file */
cmdin(file, command)
char *file, *command;
{
	struct filenode *install(), *lookup(), *p;
	struct listnode *new;
	char *ckalloc(), *strsave();

	if ((p = lookup(file)) == NULL)
		p = install(file);
	new = (struct listnode *) ckalloc(sizeof(struct listnode));
	new->string = strsave(command);
	new->next = p->cmds;
	p->cmds = new;
	++(p->cmd_nbr);
}

/* cmdout - return file's n-th command */
char *cmdout(file, n)
char *file;
int n;
{
	struct filenode *lookup(), *p;
	struct listnode *scan;
	int steps;

	if ((p = lookup(file)) == NULL || p->cmds == NULL || p->cmd_nbr < n)
		return(NULL);
	steps = p->cmd_nbr - n;
	for (scan = p->cmds; steps-- > 0; scan = scan->next)
		;
	return(scan->string);
}

/* prein - record a precursor of a file */
prein(file, precursor)
char *file, *precursor;
{
	struct filenode *install(), *lookup(), *p;
	struct listnode *new;
	char *ckalloc(), *strsave();

	if ((p = lookup(file)) == NULL)
		p = install(file);
	new = (struct listnode *) ckalloc(sizeof(struct listnode));
	new->string = strsave(precursor);
	new->next = p->pres;
	p->pres = new;
	++(p->pre_nbr);
}

/* preout - return file's n-th precursor */
char *preout(file, n)
char *file;
int n;
{
	struct filenode *lookup(), *p;
	struct listnode *scan;
	int steps;

	if ((p = lookup(file)) == NULL || p->pres == NULL || p->pre_nbr < n)
		return(NULL);
	steps = p->pre_nbr - n;
	for (scan = p->pres; steps-- > 0; scan = scan->next)
		;
	return(scan->string);
}

/* statin - record the status of a file */
statin(file, status)
char *file;
int status;
{
	struct filenode *lookup(), *p;

	if ((p = lookup(file)) != NULL)
		p->status = status;
}

/* statout - return file's current status */
int statout(file)
char *file;
{
	struct filenode *lookup(), *p;

	if ((p = lookup(file)) != NULL)
		return(p->status);
	else
		return(NOT_REACHED);
}

/* topname - return the first file name to be recorded */
char *topname()
{
	return(first_file);
}

/* lookup - search the linked list of files */
static struct filenode *lookup(file)
char *file;
{
	struct filenode *p;

	for (p = begin; p != NULL; p = p->link)
		if (strsame(file, p->name))
			return(p);
	return(NULL);
}

/* install - install a file in the linked list */
static struct filenode *install(file)
char *file;
{
	struct filenode *new;
	char *ckalloc(), *strsave();

	if (first_file == NULL)
		first_file = strsave(file);
	new = (struct filenode *) ckalloc(sizeof(struct filenode));
	new->name = strsave(file);
	new->cmds = new->pres = NULL;
	new->cmd_nbr = new->pre_nbr = 0;
	new->status = NOT_REACHED;
	/* install the node at the front of the list */
	new->link = begin;
	begin = new;
	return(begin);
}
