/*
* system.c - system-dependent procedures for the update program  (UNIX version)
*
*
* Entry points:
*
*	execute(command)
*	char *command;
*		Cause the specified command to be executed.
*
*	implicit(file)
*	char *file;
*		Add any implicit dependency relations and commands for the
*		named file to update's internal record of UPDATE.
*
*	TIME_TYPE modtime(file)
*	char *file;
*		Return file's modification time (0 for nonexistent files).
*
*
* External procedure calls (not counting system-specific functions):
*
*	cmdin(file, command)				.. file depmod.c
*	char *file, *command;
*		Record a command of a file.
*
*	char *cmdout(file, n)				.. file depmod.c
*	char *file;
*	int n;
*		Return file's n-th command (NULL if there aren't n commands).
*
*	macout(line1, line2, maxchars)			.. file macmod.c
*	char *line1, *line2;
*	int maxchars;
*		Expand macros while copying line1 to line2.  At most
*		maxchars characters are copied.
*
*	prein(file, precursor)				.. file depmod.c
*	char *file, *precursor;
*		Record a precursor of a file.
*/

#include "update.h"
#include <sys/types.h>
#include <sys/stat.h>

/* execute - execute a command; if unsuccessful, remove target file and die */
execute(command)
char *command;
{
	TIME_TYPE modtime();
	char target[NAME_SIZE];

	if (system(command) != 0) {
		/* get name of file being updated */
		macout("TARGET", target, NAME_SIZE);
		/* if target exists, remove it */
		if (modtime(target) != 0) {
			fprintf(stderr, "Error.  %s deleted.\n", target);
			unlink(target);
		}
		exit(1);
	}
}

/* implicit - install implicit dependency relations and commands */
implicit(ofile)
char *ofile;
{
	TIME_TYPE modtime();
	char *bldsrc(), *cmdout(), command[CMD_SIZE], *getnam(), sfile[NAME_SIZE];

	/* if ofile is an object file and the source file exists */
	if (strlen(ofile) >= NAME_SIZE) {
		ofile[NAME_SIZE] = '\0';
		fatalf("The name `%s...' is too long.", ofile);
	}
	if (bldsrc(ofile, sfile) != NULL && modtime(sfile) != 0) {
		prein(ofile, sfile);
		/* install the default command if none is known */
		if (cmdout(ofile, 1) == NULL) {
			sprintf(command, "COMPILE %s", sfile);
			cmdin(ofile, command);
		}
	}
}

/* modtime - return file's modification time (0 for nonexistent files) */
TIME_TYPE modtime(file)
char *file;
{
	struct stat buf;
	
	return(stat(file, &buf) >= 0 ? buf.st_mtime : 0);
}

/*
* bldsrc - construct a source file name from the name of a C object file;
* return NULL for names not belonging to C object files.
*/
#define O_SUFFIX	".o"
#define S_SUFFIX	".c"
static char *bldsrc(ofile, sfile)
char *ofile, *sfile;
{
	if (!strsame(ofile + strlen(ofile) - strlen(O_SUFFIX) , O_SUFFIX))
		return(NULL);
	strcpy(sfile, ofile);
	/* strip off O_SUFFIX */
	sfile[strlen(sfile) - strlen(O_SUFFIX)] = '\0';
	strcat(sfile, S_SUFFIX);
	return(sfile);
}
