/*
* update.c - automatic file updating
*
*
* Program description:
*
*	A command line has the form:
*		update [-i] [-n] [name]
*	Update consults the dependency description file named UPDATE, and
*	applies intrinsic rules relating C language source files and object
*	files, then deduces a minimal set of commands for updating the named
*	file.  If no name is given, then update uses the first name that appears
*	in the UPDATE file.  With the -n (no execution) option the commands are
*	merely printed, not executed. With the -i (interactive) option the
*	commands are printed and the user is asked whether the commands should
*	be executed; a response beginning with 'y' triggers command execution.
*
*
* Source files:
*
*	bldint.c	- build internal form of the UPDATE file
*	depmod.c	- data structure for dependencies
*	lib.c		- library of C procedures
*	macmod.c	- macro symbol table
*	system.c	- system-dependent procedures for the update program
*	update.c	- main program and core updating procedure
*	update.h	- macro definitions for the update program
*
*
* Portability:
*
*	The macro TIME_TYPE in update.h must give the type of time stamps on
*	files.
*
*	The macro C_NAME in update.h must give the default C compile command,
*	without the source file's name.
*
*	The macro CONCURRENT in update.h must be defined as NO if commands
*	cannot be executed concurrently by update.
*
*	The functions execute(), implicit(), and modtime() in system.c must be
*	implemented in a manner appropriate for the host system.
*
*
* External procedure calls:
*
*	int bldint()					.. file bldint.c
*		Read the UPDATE file and record the information.  Return
*		a count of the syntax errors in UPDATE.
*
*	cmdin(file, command)				.. file depmod.c
*	char *file, *command;
*		Record a command for a file.
*
*	char *cmdout(file, n)				.. file depmod.c
*	char *file;
*	int n;
*		Return file's n-th command (NULL if there aren't n commands).
*
*	execute(command)				.. file system.c
*	char *command;
*		Cause the specified command to be executed.
*
*	implicit(file)					.. file system.c
*	char *file;
*		Add any implicit dependency relations and commands for the
*		named file to update's internal record of UPDATE.
*
*	macin(name, definition)				.. file macmod.c
*	char *name, *definition;
*		Record a macro name and definition.
*
*	macout(line1, line2, maxchars)			.. file macmod.c
*	char *line1, *line2;
*	int maxchars;
*		Expand macros while copying line1 to line2.  At most
*		maxchars characters are copied.
*
*	TIME_TYPE modtime(file)				.. file system.c
*	char *file;
*		Return file's modification time (0 for nonexistent files).
*
*	char *preout(file, n)				.. file depmod.c
*	char *file;
*	int n;
*		Return file's n-th precursor (NULL if there aren't n precursors).
*
*	statin(file, status)				.. file depmod.c
*	char *file;
*	int status;
*		Record the status of a file.
*
*	int statout(file)				.. file depmod.c
*	char *file;
*		Return file's current status.
*
*	char *topname()					.. file depmod.c
*		Return the first file named in UPDATE.
*/

#include "update.h"

static int did_work = NO, iflag = NO, nflag = NO, off_line;

main(argc, argv)
int argc;
char *argv[];
{
	TIME_TYPE modtime();
	int cn, i;
	char *cmdout(), *command, *name, *topname();

	savename(PROG_NAME);	/* for error messages */

	for (i = 1; i < argc && argv[i][0] == '-'; ++i)
		if (argv[i][1] == 'i')
			iflag = YES;
		else if (argv[i][1] == 'n')
			nflag = YES;
		else
			fatalf("Illegal option: %c.", argv[i][1]);
	if (i < argc-1)
		fatalf("Correct usage: %s -i -n name.", PROG_NAME);
	macin("COMPILE", C_NAME);
	if (bldint() > 0)
		fatalf("Improper %s file.", U_NAME);
	if (i == argc) {
		/* default to first name in UPDATE */
		if ((name = topname()) == NULL)
			fatalf("No file names are given in %s.", U_NAME);
	} else
		name = argv[i];
	off_line = (iflag == YES || nflag == YES || !CONCURRENT);
	update(name);
	if (!did_work)
		printf("%s is up to date\n", name);
	else if (iflag == YES) {
		printf("execute? ");
		if (getchar() == 'y')
			for (cn = 1; (command = cmdout("", cn)) != NULL; ++cn) {
				printf("%s\n", command);
				execute(command);
			}
	}
	exit(0);
}

/*
* update - recursive procedure to update a file
*
* The procedure begins by recursively updating the precursors of
* the named file.  If the file is then older than one or more of its
* precursors, or if it does not exist, then the commands associated
* with the file are executed.
*/
static int update(file)
char *file;
{
	TIME_TYPE m_time, modtime();
	int cn, pn, status;
	char cmdbuf[CMD_SIZE], *cmdout(), *command, *precursor, *preout(),
		younger[CMD_SIZE];

	implicit(file);	/* install implicit precursors and commands */

	/* check for error conditions */
	/* if file does not exist, has no precursors, and has no commands .. */
	if (modtime(file) == 0 && preout(file, 1) == NULL && cmdout(file, 1) == NULL)
		fatalf("Don't know how to update %s.", file);
	statin(file, ACTIVE);

	/* bring precursor files up to date */
	for (pn = 1; (precursor = preout(file, pn)) != NULL; ++pn)
		if (statout(precursor) == NOT_REACHED)
			update(precursor);

	/* compare modification times */
	m_time = modtime(file);
	younger[0] = '\0';
	for (pn = 1; (precursor = preout(file, pn)) != NULL; ++pn) {
		status = statout(precursor);
		if (status != ACTIVE && modtime(precursor) > m_time
		   || status == CHANGED && off_line) {
			if (strlen(younger) + strlen(precursor) + 1 >= CMD_SIZE)
				fatalf("`%s' has too many younger precursors.",
					file);
			if (younger[0] != '\0')	/* omit the first time */
				strcat(younger, " ");
			strcat(younger, precursor);
		}
	}

	/* update file if necessary */
	if ((younger[0] != '\0' || m_time == 0) && cmdout(file, 1) != NULL) {
		/* install name of file being updated */
		macin("TARGET", file);
		/* install macro of younger precursors */
		macin("YOUNGER", younger);
		for (cn = 1; (command = cmdout(file, cn)) != NULL; ++cn) {
			/* expand macros */
			macout(command, cmdbuf, CMD_SIZE);
			printf("%s\n", cmdbuf);
			if (iflag == YES)
			/* associate command with empty file name */
				cmdin("", cmdbuf);
			else if (nflag == NO)
				execute(cmdbuf);
		}
		did_work = YES;
		statin(file, CHANGED);
	} else
		statin(file, UNCHANGED);
}
